#!/bin/bash
# set -x
################################################################################
# YAP Tools: Script for Automating YAP Synthesis Experiments 
# CC-BY-NC 4.0, (c) Mario Gleirscher
#
# Note: This script is pre-configured with settings for the CSI:Cobot workcell
# example.
################################################################################

USAGE="
Usage: ${0} [-2cdehiltvwy] [-p N]
  
       This script automates some core steps of synthesis experiments 
       with YAP and further tools.  By default (-w), a design space is 
       generated (-y), followed by the verified synthesis of policies (-e),
       and the visualisation of a pre-selected policy (-v).  A few other 
       steps and options are available with the following options:

       -2     plot 2D Pareto front
       -a     perform (quantitative) analyses (only with or after -p)
       -c     generate controller code
       -d     generate information for debugging
       -e     (re-)run EvoChecker 
       -h     help
       -i     insert decision points (only with or after -y)
       -l     show logs
       -p N   (re-)run Prism and generate policy no. N
       -t     generate abstract traces
       -v     visualise Pareto front (with -e) and a specific policy (with -p)
       -w     run overall procedure (default variant of experiment including the
              generation of controller code from a particular optimal solution)
       -y     run YAP

System requirements: 
       YAP, evoChecker, PRISM, Python 3, matplotlib,
       Java/OpenJDK (11+ for EvoChecker and YAP), zathura, bash, sed, 
       gnome-terminal, POSIX file utils

Examples: 
       Run overall procedure, picking optimal solution no. 24:
       ./run.sh -w -p 24 

       Run YAP, convert MDP into pDTMC, run evoChecker, and visualise Pareto front:
       ./run.sh -y -i -e -v

Note:  Make sure that the example directly is clear of obsolete files when 
       not using -w.

This script is part of the YAP Tools.
"

################################################################################
# dataset (experiment settings)
TYP=workcell
MDL=${TYP}
# file with optimisation objectives (can be generated by YAP)
OBJ=cost-prod-risk-HC-HS-HRW
# process activity to assess
ACT=exchWrkp
# property file for quantitative policy/DTMC assessment
PROPS=workcell_pol.props
# ACT=${TYP}
TPL=workcell-ctr-tpl
# Model execution period, e.g. 50
MXP=75
# Size of the random search population, e.g. 50
POP=100
# Search iterations, e.g. 50
MEVAL=100

# defaults:
SOL=9
# IMPORTANT: manually choose from file 03-paretofronts/${OBJ}_set
# Pick the line number and subtract 2

################################################################################
# tool/installation settings (for Linux)
PDFVIEWER=zathura
PRISM=/opt/prism-4.5-linux64/bin/prism
ECDIR=/opt/EvoChecker
ECBIN=${ECDIR}/EvoChecker-1.0.1.jar
JAVABIN=/usr/lib/jvm/java-14-openjdk-amd64/bin/java
vmArgs="-Xmx3g -XX:ParallelGCThreads=1"
YAPTOOLDIR=/usr/share/yapp/tools
PRJDIR=${HOME}/yap-examples/hrc2

FRONT=03-paretofronts/${OBJ}_front
SET=03-paretofronts/${OBJ}_set

usage() {
    echo "${USAGE}" 1>&2
    exit 1
}

error() {
    echo "Error code ${?}"
    exit 1
}

# defaults for CLI parameters
DOANA=0
DOYAP=0
DOYAPC=0
DOEVO=0
DOIDP=0
DOLOG=0
DOPRISM=0
DOVIS=0
DOTRACES=0
DETAIL=0
DIM=3

# check and initialize parameters
while getopts "2acdelp:tvwhiy" OPTION; do
    case ${OPTION} in
        2) DIM=2
           ;;
	a) DOANA=1
	   ;;
	c) DOYAPC=1
	   ;;
	e) DOEVO=1
	   ;;
	i) DOIDP=1
	   ;;
	l) DOLOG=1
	   ;;
	p) DOPRISM=1
           SOL=${OPTARG}
	   ;;
	v) DOVIS=1
	   ;;
        w) DOYAP=1
           DOEVO=1
           DOPRISM=1
           DOVIS=1
           DOYAPC=1
           ;;
	t) DOTRACES=1
	   ;;
        y) DOYAP=1
           ;;
	*) usage
	   ;;
    esac
done
# shift $((OPTIND-1))

if [[ $# < 1 ]]; then
    DOYAP=1
    DOEVO=1
    DOPRISM=1
    DOVIS=1
    DOYAPC=1
fi

################################################################################
if [[ ${DOLOG} == 1 ]]; then
    echo "Show log file(s) ..."
    tail -n 200 ~/.yap/yap.log
    exit 0
fi

################################################################################
if [[ ${DOYAP} == 1 ]]; then 
    echo "Generate controller design space using YAP ..."
    echo "(a) pDTMC (not PRISM-compatible with mitigation decision points):"
    sed -e '/YAP#EVOLVABLES_FIX%\|YAP#PREDICATES%/d' <${PRJDIR}/01-process/${MDL}-tpl.pm \
        >${PRJDIR}/01-process/${MDL}-tpl-pdtmc.pm
    yapp -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -t "${PRJDIR}/01-process/${MDL}-tpl-pdtmc.pm" \
         -o "${PRJDIR}/02-designspace/${MDL}.pm" \
         --global-logging -l 2 -f evochecker -d turn-based \
         --expand-formulas --simulate no --synthesise controller \
        || { echo "Yapp failed." && exit; };
    # --controller-optional 
    
    echo "(b) MDP (PRISM-compatible with non-deterministic controller):"
    sed -e '/^dtmc/s/dtmc/mdp/; /evolve/d; /^\/\/ const/s/^\/\/ //; /EVOLVABLES%/d' \
        <${PRJDIR}/01-process/${MDL}-tpl.pm \
        >${PRJDIR}/01-process/${MDL}-tpl-mdp.pm
    yapp -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -t "${PRJDIR}/01-process/${MDL}-tpl-mdp.pm" \
         -o "${PRJDIR}/02-designspace/${MDL}-mdp-compact.pm" \
         --global-logging -l 2 -f evochecker -d turn-based \
         --simulate no --synthesise controller --non-deterministic \
        || { echo "Yapp failed." && exit 1; };

    mv ${PRJDIR}/02-designspace/*.props ${PRJDIR}/02-requirements/
fi

################################################################################
if [[ ${DOIDP} == 1 ]]; then
    echo "Unfolding MDP with PRISM ..."
    ${PRISM} "${PRJDIR}/02-designspace/${MDL}-mdp-compact.pm" \
	     -exportmodel 04-process-opt/${MDL}-mdp-compact.tra,sta,lab \
             -mdp >04-process-opt/${MDL}-prism.log \
        || { echo "Expansion of the MDP transition relation (using PRISM) failed." && exit 1; };

    echo "Converting MDP transition matrix into pDTMC by inserting decision points ..."
    yapp -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -t "${PRJDIR}/01-process/${MDL}-tpl-with-dps.pm" \
         -e "${PRJDIR}/04-process-opt/${MDL}-mdp-compact" \
         -o "${PRJDIR}/02-designspace/${MDL}-pdtmc-with-dps.pm" \
         --global-logging -l 2 -f pdtmcalldps -d turn-based \
         --simulate no --synthesise controller --non-deterministic \
         --expand-formulas || { echo "Yapp failed." && exit 1; };

    echo "Adjusting model execution period ..."
    sed -i "s/\\(const int T = \\)[0-9]\\{1,3\\};/\\1${MXP};/" \
        ${PRJDIR}/02-designspace/${MDL}-pdtmc-with-dps.pm \
        || { echo "Failed in adjusting model execution period." && exit 1; };
    
    echo "Resetting the initial state ..."
    # turn=2: it's the operator's turn 
    # ract=2: robot is in exchWrkp activity
    sed -i 's/\(\(turn\|ract\): \[0\.\.2\] init \)[0-9]/\12/' \
        ${PRJDIR}/02-designspace/${MDL}-pdtmc-with-dps.pm \
        || { echo "Failed in resetting initial state." && exit 1; };

    echo "Replacing normal pDTMC with decision-point enhanced pDTMC ..."
    mv ${PRJDIR}/02-designspace/${MDL}-pdtmc-with-dps.pm ${PRJDIR}/02-designspace/${MDL}.pm
fi

################################################################################
# generate EvoChecker config file (for Linux)
ECCFG="
PROBLEM = ${TYP}
MODEL_TEMPLATE_FILE = ${PRJDIR}/02-designspace/${MDL}.pm
PROPERTIES_FILE     = ${PRJDIR}/02-requirements/${OBJ}.pctl
ALGORITHM = NSGAII
POPULATION_SIZE = ${POP}
MAX_EVALUATIONS = ${MEVAL}
PROCESSORS = 1
PLOT_PARETO_FRONT = false
PYTHON3_DIRECTORY = /usr/bin/python3
VERBOSE = true
JAVA=${JAVABIN}
"

if [[ ${DOEVO} == 1 ]]; then
    echo "Synthesise optimal controller using EvoChecker from ${PRJDIR}/02-designspace/${MDL}.pm ..."
    echo "${ECCFG}" >${ECDIR}/config.properties

    # run evochecker (add policy output? for the pareto front)
    export LD_LIBRARY_PATH=/opt/prism-4.5-linux64/lib

    echo "Running EvoChecker on '${MDL}' for objective '${OBJ}' ..."
    cd ${ECDIR}

    touch ${PRJDIR}/03-paretofronts/${MDL}-evo.log
    gnome-terminal -- sh -c "tail -f ${PRJDIR}/03-paretofronts/${MDL}-evo.log"
    
    ${JAVABIN} \
        -Dfile.encoding=UTF-8 \
        -classpath ${ECDIR}/target/classes:${HOME}/.m2/repository/org/hamcrest/hamcrest-core/1.3/hamcrest-core-1.3.jar:${HOME}/.m2/repository/org/antlr/antlr4/4.5/antlr4-4.5.jar:${HOME}/.m2/repository/org/antlr/antlr4-runtime/4.5/antlr4-runtime-4.5.jar:${HOME}/.m2/repository/org/abego/treelayout/org.abego.treelayout.core/1.0.1/org.abego.treelayout.core-1.0.1.jar:${HOME}/.m2/repository/org/antlr/antlr-runtime/3.5.2/antlr-runtime-3.5.2.jar:${HOME}/.m2/repository/org/antlr/ST4/4.0.8/ST4-4.0.8.jar:${HOME}/.m2/repository/org/mariuszgromada/math/MathParser.org-mXparser/4.4.2/MathParser.org-mXparser-4.4.2.jar:${HOME}/.m2/repository/org/uma/jmetal/4.5.2/jmetal-4.5.2.jar:${HOME}/.m2/repository/org/modelChecker/PrismAPI/1.0.16/PrismAPI-1.0.16.jar \
        evochecker.EvoChecker >"${PRJDIR}/03-paretofronts/${MDL}-evo.log" \
        || { echo "EvoChecker failed." && exit; };
    
    cd ${PRJDIR}

    cp ${ECDIR}/`sed -n -e '/Pareto Front:/ s/Pareto Front: \(.*$\)/\1/p' <03-paretofronts/${MDL}-evo.log` ${FRONT}
    cp ${ECDIR}/`sed -n -e '/Pareto Set:/ s/Pareto Set: \(.*$\)/\1/p' <03-paretofronts/${MDL}-evo.log` ${SET} 
fi

if [[ ${DOVIS} == 1 && -e "${FRONT}" ]]; then 
    echo "Visualise Pareto front ${FRONT} ..."
    # visualise policy set and Pareto front: run matplotlib
    if [[ ${DIM} == 2 ]]; then
        python3 ${ECDIR}/scripts/plotFront2D.py ${FRONT} &
    else
        python3 ${ECDIR}/scripts/plotFront3D.py ${FRONT} &
    fi
fi

################################################################################
if [[ ${DOPRISM} == 1 ]]; then 
    echo "Extract Pareto-optimal solution no. ${SOL} (for testing purposes) ..."
    NAMES=`sed -n -e '1p' <${SET}`
    VALUES=`sed -n -e "$((${SOL}+2))p" <${SET}`

    if [[ -z ${VALUES} ]]; then
	echo -n "Error: Solution ${SOL} not in range. Argument N for -p or " 
	echo "script parameter SOL must be in 1..$((`wc -l <${SET}` - 2))!"
	exit 1
    fi 
    echo -e "Chose solution:\n${NAMES}\n${VALUES}\n"

    # replace parameters
    # declare -A PARAMS
    # IFS=' '
    read -ra VALUES <<< "${VALUES}"
    i=0
    # create sed script for instantiating parameters with values
    for n in ${NAMES}; do
        # PARAMS[$n]=${VALUES[$i]}
        SEDSCR="${SEDSCR}s/evolve \(int\|double\) \(${n}\)\(.*$\)/const \1 \2 = ${VALUES[$i]};/;" 
        i=$(($i+1))
    done

    echo "Instantiate parametric stochastic process (pDTMC/pMDP->DTMC/MDP) ..."
    sed -e "${SEDSCR}" <02-designspace/${MDL}.pm >04-process-opt/${MDL}.dtmc

    # extract transition matrix from DTMC/MDP (previously instantiated by evoChecker)
    # Note/hack: switch -mdp has to be used to tell PRISM to export event labels from a DTMC model
    ${PRISM} 04-process-opt/${MDL}.dtmc -exportmodel 04-process-opt/${MDL}.tra,sta,lab \
             -mdp >04-process-opt/${MDL}-prism.log
    # ${PRISM} 01-process/${MDL}.pm -exportmodel 04-process-opt/${MDL}.tra,sta,lab -mdp
fi

if [[ ${DOANA} == 1 ]]; then
    echo "Perform (quantitative) analyses of the selected policy (DTMC) ..."
    if [[ -f "04-process-opt/${MDL}.dtmc" ]]; then
        rm 03-analysis/${MDL}-prism.log
        for p in TASK_FINISH5 COUNT_HARZARDS \
                 MISHAP_FREEDOM_MINMAX MISHAP_FREEDOM_AVG \
                 MISHAP_FREEDOM_MINMAX_FROM_HS \
                 MISHAP_FREEDOM_MINMAX_FROM_HRW \
                 MISHAP_FREEDOM_MINMAX_FROM_HC \
                 MISHAP_CNDPR_MINMAX_HS MISHAP_CNDPR_AVG_HS \
                 MISHAP_CNDPR_MINMAX_HRW MISHAP_CNDPR_AVG_HRW \
                 MISHAP_CNDPR_MINMAX_HC MISHAP_CNDPR_AVG_HC;
        do
	    ${PRISM} 04-process-opt/${MDL}.dtmc 02-requirements/${PROPS} \
                     -property "${p}" \
                     >>03-analysis/${MDL}-prism.log
        done 
        sed -i -n -e '/\(Model checking:\|Result:\)/p' 03-analysis/${MDL}-prism.log
    else
	echo "File 04-process-opt/${MDL}.dtmc missing. Please, use $0 -p ... first!"
    fi
fi

if [[ ${DOVIS} == 1 ]]; then
    if [[ -e 04-process-opt/${MDL}.sta
        && -e 04-process-opt/${MDL}.lab && -e 04-process-opt/${MDL}.tra ]]; then 
        echo "Visualise selected policy (DTMC) ..."
        ${YAPTOOLDIR}/yap-debug.sh -o 05-controller/ --vis-ctr 04-process-opt/${MDL} --adv2dot # --mdp
        mv 04-process-opt/${MDL}.pdf 04-process-opt/${MDL}-proc.pdf
        ${PDFVIEWER} 04-process-opt/${MDL}-proc.pdf &
        echo "Visualise controller part of the policy ..."
        ${YAPTOOLDIR}/yap-debug.sh -o 05-controller/ --vis-ctr 04-process-opt/${MDL} --ctr2dot
        mv 04-process-opt/${MDL}.pdf 04-process-opt/${MDL}-ctr.pdf
        ${PDFVIEWER} 04-process-opt/${MDL}-ctr.pdf &
    else
        echo "No policy ${MDL} to visualise. Please, provide .sta/tra/lab files."
    fi
fi

################################################################################
# extract and generate safety controller implementation (with YAP)

if [[ ${DOYAPC} == 1 ]]; then
    echo "Generate controller executable (DFA) ..."
    yapp -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -t "${PRJDIR}/05-controller/${TPL}.cs" \
         -e "${PRJDIR}/04-process-opt/${MDL}" \
         -o "${PRJDIR}/05-controller/${MDL}.cs" \
         --global-logging -l 2 -f cobotsdt \
         --simulate no --synthesise platformcode || { echo "Yapp failed." && exit; };
    # --interface .yap 
fi

################################################################################
# generate abstract traces for developing use/test cases for run-time verification
if [[ ${DOTRACES} == 1 ]]; then
    echo "Obtain abstract trace for use case-based testing from model (pMDP) ..."
    ${PRISM} 04-process-opt/${MDL}.dtmc 02-requirements/genusecases.props \
             -mdp -v -extraddinfo >04-process-opt/${MDL}-prism.log
    # Note: Unfortunately, PRISM won't export the action labels, such that
    # the simulator GUI has to be used and the witnesses for the use cases be exported.
fi
